@@ -1,7 +1,7 @@
 /*
  * Author:      William Chia-Wei Cheng (bill.cheng@acm.org)
  *
- * Copyright (C) 2001, William Chia-Wei Cheng.
+ * Copyright (C) 2001-2009, William Chia-Wei Cheng.
  *
  * This file may be distributed under the terms of the Q Public License
  * as defined by Trolltech AS of Norway and appearing in the file
@@ -15,7 +15,7 @@
  * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
- * @(#)$Header: /mm/src/tgif/v4/RCS/auxtext.c,v 4.33 2001/08/06 21:38:55 william Exp $
+ * @(#)$Header: /mm2/home/cvs/bc-src/tgif/auxtext.c,v 1.9 2009/09/30 14:53:55 william Exp $
  */
 
 #define _INCLUDE_FROM_AUXTEXT_C_
@@ -80,6 +80,25 @@ int GetNumberOfMiniLines(minilines)
    return count;
 }
 
+int IsHex(ch, pn_val)
+   char ch;
+   int *pn_val;
+{
+   if (ch >= '0' && ch <= '9') {
+      *pn_val = ((int)(ch - '0'));
+      return TRUE;
+   }
+   if (ch >= 'a' && ch <= 'f') {
+      *pn_val = 10+((int)(ch - 'a'));
+      return TRUE;
+   }
+   if (ch >= 'A' && ch <= 'F') {
+      *pn_val = 10+((int)(ch - 'F'));
+      return TRUE;
+   }
+   return FALSE;
+}
+
 int InputOctalString(pszBuf, pnBufSize)
    char *pszBuf;
    int *pnBufSize;
@@ -100,61 +119,85 @@ int InputOctalString(pszBuf, pnBufSize)
    UtilTrimBlanks(szSpec);
    if (*szSpec == '\0') return FALSE;
 
-   for (i=0, psz=szSpec; !nError && i < nBufSizeIn-1 && *psz != '\0';
-         i++, psz++) {
-      int nThisIsOctal=FALSE;
+   if (szSpec[0] == '0' && szSpec[1] == 'x') {
+      for (i=0, psz=(&szSpec[2]); !nError && i < nBufSizeIn-1 && *psz != '\0';
+            i++, psz=(&psz[2])) {
+         int hi_nibble=0, lo_nibble=0;
 
-      if (*psz == '\\') {
-         if (psz[1] == '\0') {
-            pszBuf[i] = '\\';
-         } else if (psz[1] == '\\') {
-            pszBuf[i] = '\\';
-            psz++;
-         } else if (psz[1] == 'r') {
-            pszBuf[i] = '\r';
-            psz++;
-         } else if (psz[1] == 'n') {
-            pszBuf[i] = '\n';
-            psz++;
-         } else if (psz[1] >= '0' && psz[1] <= '7') {
-            char *psz1=NULL, saved_ch='\0';
-            int oval=0;
+         if (IsHex(psz[0], &hi_nibble) && IsHex(psz[1], &lo_nibble)) {
+            int hval=(((hi_nibble&0x0ff)<<4)+(lo_nibble&0x0ff));
 
-            *psz = '0';
-            for (psz1=(&psz[2]); *psz1 >= '0' && *psz1 <= '7'; psz1++) ;
-            saved_ch = (*psz1);
-            *psz1 = '\0';
-            if (sscanf(psz, "%o", &oval) != 1) {
-               nError = TRUE;
-            } else {
-               pszBuf[i] = (char)(oval & 0xff);
+            pszBuf[i] = hval;
 
-               nThisIsOctal = TRUE;
-               if (canvasFontDoubleByte && canvasFontDoubleByteModBytes) {
-                  if (nPrevIsOctal) {
-                     nByte1OfOctal = !nByte1OfOctal;
-                  } else {
-                     nByte1OfOctal = TRUE;
-                  }
-                  if ((pszBuf[i] & 0x80) == 0) {
-                     pszBuf[i] |= 0x80;
-                  }
+            if (canvasFontDoubleByte && canvasFontDoubleByteModBytes) {
+               if ((pszBuf[i] & 0x80) == 0) {
+                  pszBuf[i] |= 0x80;
                }
             }
-            *psz1 = saved_ch;
-            *psz = '\\';
-            if (!nError) {
-               psz = psz1;
-               psz--;
+         } else {
+            nError = TRUE;
+         }
+      }
+   } else {
+      for (i=0, psz=szSpec; !nError && i < nBufSizeIn-1 && *psz != '\0';
+            i++, psz++) {
+         int nThisIsOctal=FALSE;
+
+         if (*psz == '\\') {
+            if (psz[1] == '\0') {
+               pszBuf[i] = '\\';
+            } else if (psz[1] == '\\') {
+               pszBuf[i] = '\\';
+               psz++;
+            } else if (psz[1] == 'r') {
+               pszBuf[i] = '\r';
+               psz++;
+            } else if (psz[1] == 'n') {
+               pszBuf[i] = '\n';
+               psz++;
+            } else if (psz[1] >= '0' && psz[1] <= '7') {
+               char *psz1=NULL, saved_ch='\0';
+               int oval=0;
+
+               *psz = '0';
+               for (psz1=(&psz[2]); *psz1 >= '0' && *psz1 <= '7'; psz1++) ;
+               saved_ch = (*psz1);
+               *psz1 = '\0';
+               if (sscanf(psz, "%o", &oval) != 1) {
+                  nError = TRUE;
+               } else {
+                  pszBuf[i] = (char)(oval & 0xff);
+
+                  nThisIsOctal = TRUE;
+                  if (canvasFontDoubleByte && canvasFontDoubleByteModBytes) {
+                     if (nPrevIsOctal) {
+                        nByte1OfOctal = !nByte1OfOctal;
+                     } else {
+                        nByte1OfOctal = TRUE;
+                     }
+                     if ((pszBuf[i] & 0x80) == 0) {
+                        pszBuf[i] |= 0x80;
+                     }
+                  }
+               }
+               *psz1 = saved_ch;
+               *psz = '\\';
+               if (!nError) {
+                  psz = psz1;
+                  psz--;
+               }
+            } else {
+               i--;
+               psz++;
             }
          } else {
-            i--;
-            psz++;
+            pszBuf[i] = *psz;
          }
-      } else {
-         pszBuf[i] = *psz;
+         nPrevIsOctal = nThisIsOctal;
       }
-      nPrevIsOctal = nThisIsOctal;
+   }
+   if (canvasFontDoubleByte && ((i & 0x1) != 0)) {
+      nError = TRUE;
    }
    if (nError) {
       *pszBuf = '\0';
@@ -210,7 +253,7 @@ int CharIsCntrlShiftINS(key_ev, key_sym)
  *
  * if (CharIsESC(key_ev, buf, key_sym, &has_ch)) {
  * } else if (CharIsCRorLF(key_ev, buf, key_sym, &has_ch)) {
- * } else if (CharIsBSorDEL(key_ev, buf, key_sym, &has_ch)) {
+ * } else if (CharIsBSorDEL(key_ev, buf, key_sym, &has_ch, FALSE)) {
  * } else if (CharIsTAB(key_ev, buf, key_sym, &has_ch)) {
  * }
  */
@@ -265,14 +308,33 @@ int CharIsDEL(key_ev, psz_buf, key_sym, pn_has_ch)
    return (*psz_buf == '\177');
 }
 
-int CharIsBSorDEL(key_ev, psz_buf, key_sym, pn_has_ch)
+int CharIsBS(key_ev, psz_buf, key_sym, pn_has_ch, left_is_BS)
    XKeyEvent *key_ev;
    char *psz_buf;
    KeySym key_sym;
-   int *pn_has_ch;
+   int *pn_has_ch, left_is_BS;
 {
-   if (key_sym==XK_Left || key_sym==XK_KP_Left || key_sym==XK_BackSpace ||
-         key_sym==XK_Delete || key_sym==XK_KP_Delete) {
+   if ((left_is_BS && (key_sym==XK_Left || key_sym==XK_KP_Left)) ||
+         key_sym==XK_BackSpace) {
+      return TRUE;
+   }
+   if (key_ev != NULL && (key_ev->state & ControlMask) && (key_sym == XK_h)) {
+      return TRUE;
+   }
+   if (pn_has_ch != NULL) {
+      return ((*pn_has_ch) && (*psz_buf == '\b'));
+   }
+   return (*psz_buf == '\b');
+}
+
+int CharIsBSorDEL(key_ev, psz_buf, key_sym, pn_has_ch, left_is_BS)
+   XKeyEvent *key_ev;
+   char *psz_buf;
+   KeySym key_sym;
+   int *pn_has_ch, left_is_BS;
+{
+   if ((left_is_BS && (key_sym==XK_Left || key_sym==XK_KP_Left)) ||
+         key_sym==XK_BackSpace || key_sym==XK_Delete || key_sym==XK_KP_Delete) {
       return TRUE;
    }
    if (key_ev != NULL && (key_ev->state & ControlMask) && (key_sym == XK_h)) {
@@ -320,6 +382,25 @@ int CharIsEnd(key_ev, psz_buf, key_sym, pn_has_ch)
    return (key_sym==XK_End || key_sym==XK_KP_End);
 }
 
+int CharIsCntrlSpace(key_ev, psz_buf, key_sym, pn_has_ch)
+   XKeyEvent *key_ev;
+   char *psz_buf;
+   KeySym key_sym;
+   int *pn_has_ch;
+{
+   int state=key_ev->state;
+
+   if (((state & ControlMask) == ControlMask) && ((state & METAMASK) == 0)) {
+      if (key_sym == XK_space || (psz_buf[0] == ' ' && psz_buf[1] == '\0')) {
+         return TRUE;
+      } else if (key_sym == XK_backslash || (psz_buf[0] == '\\' &&
+            psz_buf[1] == '\0')) {
+         return TRUE;
+      }
+   }
+   return FALSE;
+}
+
 void TranslateKeys(s, key_sym)
    char *s;
    KeySym *key_sym;
@@ -561,6 +642,11 @@ int UpdTextBBox(ObjPtr)
 #ifdef _TGIF_DBG /* debug, do not translate */
    int saved_baseline_y=text_ptr->baseline_y;
 
+#endif /* _TGIF_DBG */
+
+   if (mainDisplay == NULL) return FALSE;
+
+#ifdef _TGIF_DBG /* debug, do not translate */
    if (PRTGIF && !cmdLineOpenDisplay) {
       fprintf(stderr, "UpdTextBBox() called in PRTGIF!\n");
    }
@@ -705,6 +791,7 @@ int PaintBackingPixmap(text_ptr, bbox_w, bbox_h, min_lbearing, max_rextra,
    return TRUE;
 }
 
+#ifdef NOT_DEFINED
 #ifdef _TGIF_DBG /* debug, do not translate */
 static
 void DebugCachedTextBitmapImage(image, before, w, h, min_lbearing, max_rextra)
@@ -731,6 +818,7 @@ void DebugCachedTextBitmapImage(image, before, w, h, min_lbearing, max_rextra)
    }
 }
 #endif /* _TGIF_DBG */
+#endif /* NOT_DEFINED */
 
 void MakeCachedTextBitmap(ObjPtr)
    struct ObjRec *ObjPtr;
@@ -1347,7 +1435,7 @@ void DumpTextFill(FP, ObjPtr)
          if (preDumpSetup) PSUseColorPattern();
          DumpRectPath(FP, ltx, lty, rbx, rby, 3, FALSE);
          fprintf(FP, "%s %s\n", gPsCmd[PS_EOCLIP], gPsCmd[PS_NEWPATH]);
-         DumpPatFill(FP, fill, 8, ObjPtr->bbox, 3, TRUE);
+         DumpPatFill(FP, fill, ObjPtr->bbox, 3, TRUE);
       } else {
          DumpRectPath(FP, ltx, lty, rbx, rby, 3, FALSE);
          fprintf(FP, "%s\n", gPsCmd[PS_FILL]);
@@ -1795,6 +1883,28 @@ void SaveTextObj(FP, ObjPtr)
    if (fprintf(FP, "])") == EOF) writeFileFailed = TRUE;
 }
 
+void SaveHexString(FP, buf)
+   FILE *FP;
+   char *buf;
+{
+   for ( ; *buf != '\0'; buf++) {
+      int ival=0;
+
+      ival = ((((unsigned int)(*buf)) >> 4) & 0x0f);
+      if (ival >= 10) {
+         fprintf(FP, "%c", (char)(((int)'a')+(ival-10)));
+      } else {
+         fprintf(FP, "%c", (char)(((int)'0')+ival));
+      }
+      ival = (((unsigned int)(*buf)) & 0x0f);
+      if (ival >= 10) {
+         fprintf(FP, "%c", (char)(((int)'a')+(ival-10)));
+      } else {
+         fprintf(FP, "%c", (char)(((int)'0')+ival));
+      }
+   }
+}
+
 char *ReadString(Str)
    char *Str;
 {
@@ -1846,20 +1956,23 @@ StrSegInfo *CreateStrSegFromString(buf, pOwnerStrBlock)
    pStrSeg->font = curFont;
    pStrSeg->style = curStyle;
    pStrSeg->sz_unit = GetCurSzUnit();
-   strcpy(pStrSeg->color_str, colorMenuItems[colorIndex]);
-
+   if (mainDisplay != NULL) {
+      strcpy(pStrSeg->color_str, colorMenuItems[colorIndex]);
+   }
    pStrSeg->double_byte = canvasFontDoubleByte;
    pStrSeg->double_byte_mod_bytes = canvasFontDoubleByteModBytes;
    pStrSeg->double_byte_vertical = canvasFontDoubleByteVertical;
    pStrSeg->direction = canvasFontDirection;
    pStrSeg->dontreencode = canvasFontDontReencode;
    pStrSeg->underline_on = curUnderlineOn;
+   pStrSeg->overline_on = curOverlineOn;
 
    pStrSeg->asc = canvasFontAsc;
    pStrSeg->des = canvasFontDes;
-   pStrSeg->w = MyTextWidth(canvasFontPtr, pStrSeg->dyn_str.s,
-         pStrSeg->dyn_str.sz-1);
-
+   if (mainDisplay != NULL) {
+      pStrSeg->w = MyTextWidth(canvasFontPtr, pStrSeg->dyn_str.s,
+            pStrSeg->dyn_str.sz-1);
+   }
    pStrSeg->owner = pOwnerStrBlock;
 
    return pStrSeg;
@@ -1976,6 +2089,10 @@ struct ObjRec *FormTextObjectFromFile(FP, AbsX, AbsBaselineY)
    obj_ptr->y = AbsBaselineY-text_ptr->minilines.first->asc;
    obj_ptr->type = OBJ_TEXT;
    obj_ptr->color = colorIndex;
+   if (mainDisplay != NULL) {
+      UtilStrCpyN(obj_ptr->color_str, sizeof(obj_ptr->color_str),
+            colorMenuItems[colorIndex]);
+   }
    obj_ptr->id = objId++;;
    obj_ptr->dirty = FALSE;
    obj_ptr->rotation = 0;
