@@ -1,7 +1,7 @@
 /*
  * Author:      William Chia-Wei Cheng (bill.cheng@acm.org)
  *
- * Copyright (C) 2001, William Chia-Wei Cheng.
+ * Copyright (C) 2001-2009, William Chia-Wei Cheng.
  *
  * This file may be distributed under the terms of the Q Public License
  * as defined by Trolltech AS of Norway and appearing in the file
@@ -15,7 +15,7 @@
  * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
- * @(#)$Header: /mm/src/tgif/v4/RCS/color.c,v 4.41 2001/08/06 21:38:58 william Exp $
+ * @(#)$Header: /mm2/home/cvs/bc-src/tgif/color.c,v 1.16 2009/08/13 01:14:20 cvsps Exp $
  */
 
 #define _INCLUDE_FROM_COLOR_C_
@@ -46,8 +46,10 @@
 #include "rect.e"
 #include "select.e"
 #include "setup.e"
+#include "stk.e"
 #include "strtbl.e"
 #include "text.e"
+#include "tgcwheel.e"
 #include "util.e"
 #include "xpixmap.e"
 
@@ -196,7 +198,6 @@ int TgifParseColor(psz_color, p_color)
    return (int)XParseColor(mainDisplay, mainColormap, psz_color, p_color);
 }
 
-static
 int ParseAndAllocColorByName(colorname, color, red_req, green_req, blue_req)
    char *colorname;
    XColor *color;
@@ -1228,6 +1229,10 @@ int ChangeObjColor(ObjPtr, ColorIndex)
          }
       } else if (ObjPtr->color != ColorIndex) {
          ObjPtr->color = ColorIndex;
+         if (mainDisplay != NULL) {
+            UtilStrCpyN(ObjPtr->color_str, sizeof(ObjPtr->color_str),
+                  colorMenuItems[ObjPtr->color]);
+         }
          if (ObjPtr->type == OBJ_XPM) {
             /* RecolorXPmObj(ObjPtr, ColorIndex); */
          }
@@ -1271,8 +1276,9 @@ int ChangeObjColor(ObjPtr, ColorIndex)
    if ((canChangeAttrColor || changePropertiesOfAttrs) &&
          ObjPtr->type != OBJ_TEXT) {
       for (attr_ptr=ObjPtr->fattr; attr_ptr != NULL; attr_ptr=attr_ptr->next) {
-         if (attr_ptr->obj->color != ColorIndex) {
-            attr_ptr->obj->color = ColorIndex;
+         if (SetMiniLinesProperty(PROP_MASK_COLOR, ColorIndex,
+               &attr_ptr->obj->detail.t->minilines)) {
+            InvalidateTextCache(attr_ptr->obj->detail.t);
             changed = TRUE;
          }
       }
@@ -1541,7 +1547,7 @@ void CleanUpColors()
    defaultColorIndex = 9;
    colorIndex = 0;
 
-   if (upend_bits) { }
+   if (upend_bits == NULL) { }
 }
 
 static
@@ -1783,7 +1789,6 @@ void ColorDummyEventHandler(input)
 
 typedef struct tagUpdatePixelRec {
    int picture_changed;
-   char **old_color_str;
 } UpdatePixelInfo;
 
 static
@@ -1793,25 +1798,27 @@ void UpdatePixelForStrSeg(pStrSeg, pUserData)
 {
    UpdatePixelInfo *pupi=((UpdatePixelInfo*)pUserData);
    int new_alloc=FALSE;
-   int index=QuickFindColorIndex(NULL, pupi->old_color_str[pStrSeg->color],
-         &new_alloc, TRUE);
+   int index=QuickFindColorIndex(NULL, pStrSeg->color_str, &new_alloc, TRUE);
 
    if (index != pStrSeg->color) pupi->picture_changed = TRUE;
    if (index == INVALID) {
       sprintf(gszMsgBox,
             TgLoadCachedString(CSTID_CANNOT_ALLOC_COLOR_USE_ALT),
-            pupi->old_color_str[pStrSeg->color], colorMenuItems[colorIndex]);
+            pStrSeg->color_str, colorMenuItems[colorIndex]);
       Msg(gszMsgBox);
       allocColorFailed = TRUE;
       pStrSeg->color = colorIndex;
    } else {
       pStrSeg->color = index;
    }
+   if (mainDisplay != NULL) {
+      UtilStrCpyN(pStrSeg->color_str, sizeof(pStrSeg->color_str),
+            colorMenuItems[pStrSeg->color]);
+   }
 }
 
-int UpdatePixel(ObjPtr, OldColorStr)
+int UpdatePixel(ObjPtr)
    struct ObjRec *ObjPtr;
-   char **OldColorStr;
 {
    int c, i, r;
    int ncolors=0, new_alloc=FALSE, index=0, changed=FALSE, len=0;
@@ -1828,25 +1835,29 @@ int UpdatePixel(ObjPtr, OldColorStr)
    case OBJ_PIN:
       for (obj_ptr=ObjPtr->detail.r->last; obj_ptr!=NULL;
             obj_ptr=obj_ptr->prev) {
-         if (UpdatePixel(obj_ptr, OldColorStr)) {
+         if (UpdatePixel(obj_ptr)) {
             picture_changed = TRUE;
          }
       }
       break;
    case OBJ_XPM:
-      index = QuickFindColorIndex(NULL, OldColorStr[ObjPtr->color],
+      index = QuickFindColorIndex(NULL, ObjPtr->color_str,
             &new_alloc, TRUE);
       if (index != ObjPtr->color) picture_changed = TRUE;
       if (index == INVALID) {
          sprintf(gszMsgBox,
                TgLoadCachedString(CSTID_CANNOT_ALLOC_COLOR_USE_ALT),
-               OldColorStr[ObjPtr->color], colorMenuItems[colorIndex]);
+               ObjPtr->color_str, colorMenuItems[colorIndex]);
          Msg(gszMsgBox);
          allocColorFailed = TRUE;
          ObjPtr->color = colorIndex;
       } else {
          ObjPtr->color = index;
       }
+      if (mainDisplay != NULL) {
+         UtilStrCpyN(ObjPtr->color_str, sizeof(ObjPtr->color_str),
+               colorMenuItems[ObjPtr->color]);
+      }
       changed = FALSE;
       xpm_ptr = ObjPtr->detail.xpm;
       ncolors = xpm_ptr->ncolors;
@@ -1855,27 +1866,31 @@ int UpdatePixel(ObjPtr, OldColorStr)
       if (from_pixels == NULL || to_pixels == NULL) FailAllocMessage();
       for (i=0; i < ncolors; i++) {
          from_pixels[i] = xpm_ptr->pixels[i];
-         index = QuickFindColorIndex(NULL, xpm_ptr->color_str[i],
-               &new_alloc, TRUE);
-         if (index == INVALID) {
-            sprintf(gszMsgBox,
-                  TgLoadCachedString(CSTID_CANNOT_ALLOC_COLOR_USE_ALT),
-                  xpm_ptr->color_str[i], colorMenuItems[colorIndex]);
-            Msg(gszMsgBox);
-            allocColorFailed = TRUE;
-
-            len = strlen(colorMenuItems[colorIndex]);
-            free(xpm_ptr->color_str[i]);
-            xpm_ptr->color_str[i] = (char*)malloc((len+1)*sizeof(char));
-            if (xpm_ptr->color_str[i] == NULL) FailAllocMessage();
-            strcpy(xpm_ptr->color_str[i], colorMenuItems[colorIndex]);
-
-            xpm_ptr->pixels[i] = colorPixels[colorIndex];
+         if (xpm_ptr->pixels[i] == (-1)) {
+            /* transparent pixel */
          } else {
-            if (xpm_ptr->pixels[i] != colorPixels[index]) {
-               changed = TRUE;
+            index = QuickFindColorIndex(NULL, xpm_ptr->color_str[i],
+                  &new_alloc, TRUE);
+            if (index == INVALID) {
+               sprintf(gszMsgBox,
+                     TgLoadCachedString(CSTID_CANNOT_ALLOC_COLOR_USE_ALT),
+                     xpm_ptr->color_str[i], colorMenuItems[colorIndex]);
+               Msg(gszMsgBox);
+               allocColorFailed = TRUE;
+
+               len = strlen(colorMenuItems[colorIndex]);
+               free(xpm_ptr->color_str[i]);
+               xpm_ptr->color_str[i] = (char*)malloc((len+1)*sizeof(char));
+               if (xpm_ptr->color_str[i] == NULL) FailAllocMessage();
+               strcpy(xpm_ptr->color_str[i], colorMenuItems[colorIndex]);
+
+               xpm_ptr->pixels[i] = colorPixels[colorIndex];
+            } else {
+               if (xpm_ptr->pixels[i] != colorPixels[index]) {
+                  changed = TRUE;
+               }
+               xpm_ptr->pixels[i] = colorPixels[index];
             }
-            xpm_ptr->pixels[i] = colorPixels[index];
          }
          to_pixels[i] = xpm_ptr->pixels[i];
       }
@@ -1917,25 +1932,28 @@ int UpdatePixel(ObjPtr, OldColorStr)
       free(to_pixels);
       break;
    default:
-      index = QuickFindColorIndex(NULL, OldColorStr[ObjPtr->color],
+      index = QuickFindColorIndex(NULL, ObjPtr->color_str,
             &new_alloc, TRUE);
       if (index != ObjPtr->color) picture_changed = TRUE;
       if (index == INVALID) {
          sprintf(gszMsgBox,
                TgLoadCachedString(CSTID_CANNOT_ALLOC_COLOR_USE_ALT),
-               OldColorStr[ObjPtr->color], colorMenuItems[colorIndex]);
+               ObjPtr->color_str, colorMenuItems[colorIndex]);
          Msg(gszMsgBox);
          allocColorFailed = TRUE;
          ObjPtr->color = colorIndex;
       } else {
          ObjPtr->color = index;
       }
+      if (mainDisplay != NULL) {
+         UtilStrCpyN(ObjPtr->color_str, sizeof(ObjPtr->color_str),
+               colorMenuItems[ObjPtr->color]);
+      }
       if (ObjPtr->type == OBJ_TEXT) {
          UpdatePixelInfo upi;
 
          memset(&upi, 0, sizeof(UpdatePixelInfo));
          upi.picture_changed = FALSE;
-         upi.old_color_str = OldColorStr;
          DoFuncOnStrSegForMiniLines(&ObjPtr->detail.t->minilines,
                UpdatePixelForStrSeg, &upi);
          if (upi.picture_changed) {
@@ -1945,7 +1963,7 @@ int UpdatePixel(ObjPtr, OldColorStr)
       break;
    }
    for (attr_ptr=ObjPtr->fattr; attr_ptr != NULL; attr_ptr=attr_ptr->next) {
-      if (UpdatePixel(attr_ptr->obj, OldColorStr)) {
+      if (UpdatePixel(attr_ptr->obj)) {
          picture_changed = TRUE;
       }
    }
@@ -1980,10 +1998,8 @@ void UpdateXPmObjects(ObjPtr)
 
 int FlushColormap()
 {
-   register int i;
-   int old_allocated_colors, changed=FALSE;
-   char **old_color_str;
-   struct ObjRec *obj_ptr;
+   int changed=FALSE;
+   struct ObjRec *obj_ptr=NULL;
 
 #ifdef DONT_FREE_COLORMAP
    Msg(TgLoadString(STID_COLORMAP_NOT_RESETED));
@@ -1999,21 +2015,13 @@ int FlushColormap()
    if (newColormapUsed) {
       struct PageRec *page_ptr=NULL;
       int saved_color_layers=FALSE;
+      struct AttrRec *attr_ptr=NULL;
 
       for (page_ptr=firstPage; page_ptr != NULL; page_ptr=page_ptr->next) {
          for (obj_ptr=page_ptr->bot; obj_ptr!=NULL; obj_ptr=obj_ptr->prev) {
             UpdateXPmObjects(obj_ptr);
          }
       }
-      old_color_str = (char**)malloc(maxColors*sizeof(char*));
-      if (old_color_str == NULL) FailAllocMessage();
-      old_allocated_colors = maxColors;
-      for (i=0; i < maxColors; i++) {
-         old_color_str[i] =
-               (char*)malloc((strlen(colorMenuItems[i])+1)*sizeof(char));
-         if (old_color_str[i] == NULL) FailAllocMessage();
-         strcpy(old_color_str[i], colorMenuItems[i]);
-      }
       initColorDontReload = TRUE;
       CleanUpColors();
       XFreeColormap(mainDisplay, mainColormap);
@@ -2066,14 +2074,20 @@ int FlushColormap()
       gnUpdatePixelObjCount = 0;
       for (page_ptr=firstPage; page_ptr != NULL; page_ptr=page_ptr->next) {
          for (obj_ptr=page_ptr->bot; obj_ptr!=NULL; obj_ptr=obj_ptr->prev) {
-            if (UpdatePixel(obj_ptr, old_color_str)) {
+            if (UpdatePixel(obj_ptr)) {
+               changed = TRUE;
+            }
+         }
+      }
+      if (tgifObj->fattr != NULL) {
+         for (attr_ptr=tgifObj->fattr; attr_ptr != NULL;
+               attr_ptr=attr_ptr->next) {
+            if (UpdatePixel(attr_ptr->obj)) {
                changed = TRUE;
             }
          }
       }
       RestoreStatusStrings();
-      for (i=0; i < old_allocated_colors; i++) free(old_color_str[i]);
-      free(old_color_str);
 
       DestroyPinnedMenu(MENU_COLOR);
       if (colorLayers) {
@@ -2115,6 +2129,11 @@ void AddColor()
    }
 }
 
+void ChooseColor()
+{
+   TgColorWheel_HandleCntrlSpace(mainDisplay, drawWindow);
+}
+
 void SetAltEditTextBgColor()
 {
    char spec[MAXSTRING+1];
