@@ -1,7 +1,7 @@
 /*
  * Author:      William Chia-Wei Cheng (bill.cheng@acm.org)
  *
- * Copyright (C) 2001-2003, William Chia-Wei Cheng.
+ * Copyright (C) 2001-2009, William Chia-Wei Cheng.
  *
  * This file may be distributed under the terms of the Q Public License
  * as defined by Trolltech AS of Norway and appearing in the file
@@ -15,7 +15,7 @@
  * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
- * @(#)$Header: /mm/src/tgif/v4/RCS/raster.c,v 4.21 2003/08/16 19:38:52 william Exp $
+ * @(#)$Header: /mm2/home/cvs/bc-src/tgif/raster.c,v 1.20 2009/06/11 04:17:18 cvsps Exp $
  */
 
 #define _INCLUDE_FROM_RASTER_C_
@@ -55,6 +55,7 @@
 #include "xbm/freehand.xbm"
 #include "xbm/vmode.xbm"
 #include "xbm/rot_mode.xbm"
+#include "xbm/rot_mode2.xbm"
 
 #include "xbm/pat0.xbm"
 #include "xbm/pat1.xbm"
@@ -158,6 +159,7 @@
 #include "xbm/lt0.xbm"
 #include "xbm/lt1.xbm"
 #include "xbm/lt2.xbm"
+#include "xbm/lt3.xbm"
 
 #include "xbm/ls0.xbm"
 #include "xbm/ls1.xbm"
@@ -185,6 +187,7 @@
 #include "xbm/lt0s.xbm"
 #include "xbm/lt1s.xbm"
 #include "xbm/lt2s.xbm"
+#include "xbm/lt3s.xbm"
 
 #include "xbm/ls0s.xbm"
 #include "xbm/ls1s.xbm"
@@ -215,6 +218,7 @@
 #include "xbm/jpeg.xbm"
 #include "xbm/pbm.xbm"
 #include "xbm/netlist.xbm"
+#include "xbm/svg.xbm"
 
 #include "xbm/file.xbm"
 #include "xbm/landscap.xbm"
@@ -284,6 +288,7 @@ GC defaultGC=NULL;
 GC revDefaultGC=NULL;
 GC revGrayGC=NULL;
 
+Pixmap rotateModePixmap[2];
 Pixmap choicePixmap[MAXCHOICES];
 Pixmap patPixmap[MAXPATTERNS+1];
 Pixmap *lineWidthPixmap=NULL;
@@ -340,6 +345,7 @@ static int pat_w[MAXPATTERNS+1], pat_h[MAXPATTERNS+1];
 static int shape_w[MAXSHAPES+1], shape_h[MAXSHAPES+1];
 
 static unsigned char *patData[MAXPATTERNS+1];
+static int needToFreePat[MAXPATTERNS+1];
 static unsigned char *shapeData[MAXSHAPES+1];
 
 static int newLineWidthAllocated=FALSE;
@@ -554,6 +560,108 @@ void InitLines()
    }
 }
 
+static
+void SetCustomPatDir()
+{
+   *customPatDir = '\0';
+   if (*cmdLineCustomPatternDir != '\0') {
+      UtilStrCpyN(customPatDir, sizeof(customPatDir), cmdLineCustomPatternDir);
+   } else if (!PRTGIF || cmdLineOpenDisplay) {
+      char *c_ptr=NULL;
+
+      if ((c_ptr=XGetDefault(mainDisplay,TOOL_NAME,"CustomPatternDir"))!=NULL) {
+         if (c_ptr != NULL && *c_ptr != '\0' && UtilIsDirectory(c_ptr)) {
+            int len=0;
+
+            UtilStrCpyN(customPatDir, sizeof(customPatDir), c_ptr);
+            len = strlen(customPatDir);
+            if (len > 0 && customPatDir[len-1] == DIR_SEP) {
+               customPatDir[len-1] = '\0';
+            }
+         }
+      }
+   }
+}
+
+void InitPatFill()
+{
+   register int i;
+
+   pat_w[0] = pat0_width; pat_h[0] = pat0_height; patData[0] = pat0_bits;
+   pat_w[1] = pat1_width; pat_h[1] = pat1_height; patData[1] = pat1_bits;
+   pat_w[2] = pat2_width; pat_h[2] = pat2_height; patData[2] = pat2_bits;
+   pat_w[3] = pat3_width; pat_h[3] = pat3_height; patData[3] = pat3_bits;
+   pat_w[4] = pat4_width; pat_h[4] = pat4_height; patData[4] = pat4_bits;
+   pat_w[5] = pat5_width; pat_h[5] = pat5_height; patData[5] = pat5_bits;
+   pat_w[6] = pat6_width; pat_h[6] = pat6_height; patData[6] = pat6_bits;
+   pat_w[7] = pat7_width; pat_h[7] = pat7_height; patData[7] = pat7_bits;
+   pat_w[8] = pat8_width; pat_h[8] = pat8_height; patData[8] = pat8_bits;
+   pat_w[9] = pat9_width; pat_h[9] = pat9_height; patData[9] = pat9_bits;
+   pat_w[10] = pat10_width; pat_h[10] = pat10_height; patData[10] = pat10_bits;
+   pat_w[11] = pat11_width; pat_h[11] = pat11_height; patData[11] = pat11_bits;
+   pat_w[12] = pat12_width; pat_h[12] = pat12_height; patData[12] = pat12_bits;
+   pat_w[13] = pat13_width; pat_h[13] = pat13_height; patData[13] = pat13_bits;
+   pat_w[14] = pat14_width; pat_h[14] = pat14_height; patData[14] = pat14_bits;
+   pat_w[15] = pat15_width; pat_h[15] = pat15_height; patData[15] = pat15_bits;
+   pat_w[16] = pat16_width; pat_h[16] = pat16_height; patData[16] = pat16_bits;
+   pat_w[17] = pat17_width; pat_h[17] = pat17_height; patData[17] = pat17_bits;
+   pat_w[18] = pat18_width; pat_h[18] = pat18_height; patData[18] = pat18_bits;
+   pat_w[19] = pat19_width; pat_h[19] = pat19_height; patData[19] = pat19_bits;
+   pat_w[20] = pat20_width; pat_h[20] = pat20_height; patData[20] = pat20_bits;
+   pat_w[21] = pat21_width; pat_h[21] = pat21_height; patData[21] = pat21_bits;
+   pat_w[22] = pat22_width; pat_h[22] = pat22_height; patData[22] = pat22_bits;
+   pat_w[23] = pat23_width; pat_h[23] = pat23_height; patData[23] = pat23_bits;
+   pat_w[24] = pat24_width; pat_h[24] = pat24_height; patData[24] = pat24_bits;
+   pat_w[25] = pat25_width; pat_h[25] = pat25_height; patData[25] = pat25_bits;
+   pat_w[26] = pat26_width; pat_h[26] = pat26_height; patData[26] = pat26_bits;
+   pat_w[27] = pat27_width; pat_h[27] = pat27_height; patData[27] = pat27_bits;
+   pat_w[28] = pat28_width; pat_h[28] = pat28_height; patData[28] = pat28_bits;
+   pat_w[29] = pat29_width; pat_h[29] = pat29_height; patData[29] = pat29_bits;
+   pat_w[30] = pat30_width; pat_h[30] = pat30_height; patData[30] = pat30_bits;
+   pat_w[31] = pat31_width; pat_h[31] = pat31_height; patData[31] = pat31_bits;
+   pat_w[32] = pat32_width; pat_h[32] = pat32_height; patData[32] = pat32_bits;
+
+   for (i=0; i < MAXPATTERNS+1; i++) {
+      needToFreePat[i] = FALSE;
+   }
+}
+
+static int gnPatFillInitialized=FALSE;
+
+void Init2PatFill()
+{
+   register int i;
+
+   if (gnPatFillInitialized) return;
+
+   SetCustomPatDir();
+
+   for (i=3; i < MAXPATTERNS+1; i++) {
+      if (*customPatDir != '\0') {
+         unsigned int w_return=0, h_return=0;
+         int x_hot=0, y_hot=0;
+         unsigned char *data_return=NULL;
+         char path[MAXPATHLENGTH<<1];
+
+         snprintf(path, sizeof(path), "%s%cpat%1d.xbm", customPatDir, DIR_SEP,
+               i);
+         if (UtilPathExists(path)) {
+            if (XReadBitmapFileData(path, &w_return, &h_return, &data_return,
+                  &x_hot, &y_hot) == BitmapSuccess) {
+               pat_w[i] = w_return;
+               pat_h[i] = h_return;
+               patData[i] = data_return;
+               needToFreePat[i] = TRUE;
+            } else {
+               fprintf(stderr, TgLoadString(STID_INVALID_XBM_CUST_PAT), path);
+               fprintf(stderr, "\n");
+            }
+         }
+      }
+   }
+   gnPatFillInitialized = TRUE;
+}
+
 void InitPattern()
 {
    register int i;
@@ -590,40 +698,8 @@ void InitPattern()
    revDefaultGC = XCreateGC(mainDisplay, mainWindow, GCForeground |
          GCBackground | GCFunction | GCFillStyle | GCFont, &values);
 
-   pat_w[0] = pat0_width; pat_h[0] = pat0_height; patData[0] = pat0_bits;
-   pat_w[1] = pat1_width; pat_h[1] = pat1_height; patData[1] = pat1_bits;
-   pat_w[2] = pat2_width; pat_h[2] = pat2_height; patData[2] = pat2_bits;
-   pat_w[3] = pat3_width; pat_h[3] = pat3_height; patData[3] = pat3_bits;
-   pat_w[4] = pat4_width; pat_h[4] = pat4_height; patData[4] = pat4_bits;
-   pat_w[5] = pat5_width; pat_h[5] = pat5_height; patData[5] = pat5_bits;
-   pat_w[6] = pat6_width; pat_h[6] = pat6_height; patData[6] = pat6_bits;
-   pat_w[7] = pat7_width; pat_h[7] = pat7_height; patData[7] = pat7_bits;
-   pat_w[8] = pat8_width; pat_h[8] = pat8_height; patData[8] = pat8_bits;
-   pat_w[9] = pat9_width; pat_h[9] = pat9_height; patData[9] = pat9_bits;
-   pat_w[10] = pat10_width; pat_h[10] = pat10_height; patData[10] = pat10_bits;
-   pat_w[11] = pat11_width; pat_h[11] = pat11_height; patData[11] = pat11_bits;
-   pat_w[12] = pat12_width; pat_h[12] = pat12_height; patData[12] = pat12_bits;
-   pat_w[13] = pat13_width; pat_h[13] = pat13_height; patData[13] = pat13_bits;
-   pat_w[14] = pat14_width; pat_h[14] = pat14_height; patData[14] = pat14_bits;
-   pat_w[15] = pat15_width; pat_h[15] = pat15_height; patData[15] = pat15_bits;
-   pat_w[16] = pat16_width; pat_h[16] = pat16_height; patData[16] = pat16_bits;
-   pat_w[17] = pat17_width; pat_h[17] = pat17_height; patData[17] = pat17_bits;
-   pat_w[18] = pat18_width; pat_h[18] = pat18_height; patData[18] = pat18_bits;
-   pat_w[19] = pat19_width; pat_h[19] = pat19_height; patData[19] = pat19_bits;
-   pat_w[20] = pat20_width; pat_h[20] = pat20_height; patData[20] = pat20_bits;
-   pat_w[21] = pat21_width; pat_h[21] = pat21_height; patData[21] = pat21_bits;
-   pat_w[22] = pat22_width; pat_h[22] = pat22_height; patData[22] = pat22_bits;
-   pat_w[23] = pat23_width; pat_h[23] = pat23_height; patData[23] = pat23_bits;
-   pat_w[24] = pat24_width; pat_h[24] = pat24_height; patData[24] = pat24_bits;
-   pat_w[25] = pat25_width; pat_h[25] = pat25_height; patData[25] = pat25_bits;
-   pat_w[26] = pat26_width; pat_h[26] = pat26_height; patData[26] = pat26_bits;
-   pat_w[27] = pat27_width; pat_h[27] = pat27_height; patData[27] = pat27_bits;
-   pat_w[28] = pat28_width; pat_h[28] = pat28_height; patData[28] = pat28_bits;
-   pat_w[29] = pat29_width; pat_h[29] = pat29_height; patData[29] = pat29_bits;
-   pat_w[30] = pat30_width; pat_h[30] = pat30_height; patData[30] = pat30_bits;
-   pat_w[31] = pat31_width; pat_h[31] = pat31_height; patData[31] = pat31_bits;
-   pat_w[32] = pat32_width; pat_h[32] = pat32_height; patData[32] = pat32_bits;
-
+   InitPatFill();
+   Init2PatFill();
    for (i = 0; i < MAXPATTERNS+1; i++) {
       patPixmap[i] = XCreateBitmapFromData(mainDisplay, mainWindow,
             (char *)patData[i], pat_w[i], pat_h[i]);
@@ -765,6 +841,8 @@ void InitPattern()
          (char *)lt1_bits, lt1_width, lt1_height);
    lineTypePixmap[2] = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)lt2_bits, lt2_width, lt2_height);
+   lineTypePixmap[3] = XCreateBitmapFromData(mainDisplay, mainWindow,
+         (char *)lt3_bits, lt3_width, lt3_height);
 
    lineDashPixmap[0] = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)ld0_bits, ld0_width, ld0_height);
@@ -938,6 +1016,8 @@ void InitPattern()
          mainWindow, (char *)pbm_bits, pbm_width, pbm_height);
    whereToPrintPixmap[NETLIST_FILE] = XCreateBitmapFromData(mainDisplay,
          mainWindow, (char *)netlist_bits, netlist_width, netlist_height);
+   whereToPrintPixmap[SVG_FILE] = XCreateBitmapFromData(mainDisplay,
+         mainWindow, (char *)svg_bits, svg_width, svg_height);
 
    shortLineTypePixmap[0] = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)lt0s_bits, lt0s_width, lt0s_height);
@@ -945,6 +1025,8 @@ void InitPattern()
          (char *)lt1s_bits, lt1s_width, lt1s_height);
    shortLineTypePixmap[2] = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)lt2s_bits, lt2s_width, lt2s_height);
+   shortLineTypePixmap[3] = XCreateBitmapFromData(mainDisplay, mainWindow,
+         (char *)lt3s_bits, lt3s_width, lt3s_height);
 
    shortDashPixmap[0] = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)ld0s_bits, ld0s_width, ld0s_height);
@@ -980,6 +1062,11 @@ void InitPattern()
    defaultShortLineWidthPixmap[6] = XCreateBitmapFromData(mainDisplay,
          mainWindow, (char *)lw6s_bits, lw6s_width, lw6s_height);
 
+   rotateModePixmap[0] = XCreateBitmapFromData(mainDisplay, mainWindow,
+         (char *)rot_mode_bits, rot_mode_width, rot_mode_height);
+   rotateModePixmap[1] = XCreateBitmapFromData(mainDisplay, mainWindow,
+         (char *)rot_mode2_bits, rot_mode2_width, rot_mode2_height);
+
    choicePixmap[NOTHING] = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)arrow_bits, arrow_width, arrow_height);
    choicePixmap[DRAWTEXT] = XCreateBitmapFromData(mainDisplay, mainWindow,
@@ -1006,8 +1093,7 @@ void InitPattern()
          (char *)freehand_bits, freehand_width, freehand_height);
    choicePixmap[VERTEXMODE] = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)vmode_bits, vmode_width, vmode_height);
-   choicePixmap[ROTATEMODE] = XCreateBitmapFromData(mainDisplay, mainWindow,
-         (char *)rot_mode_bits, rot_mode_width, rot_mode_height);
+   choicePixmap[ROTATEMODE] = rotateModePixmap[1];
 
    wireBitmap = XCreateBitmapFromData(mainDisplay, mainWindow,
          (char *)wire_bits, wire_width, wire_height);
@@ -1086,11 +1172,28 @@ void RedrawChoiceWindow()
    ShowPageLayout();
 }
 
+void CleanUpPatFill()
+{
+   register int i;
+
+   if (!gnPatFillInitialized) return;
+
+   for (i = 0; i < MAXPATTERNS+1; i++) {
+      if (needToFreePat[i]) {
+         XFree(patData[i]);
+      }
+   }
+   gnPatFillInitialized = FALSE;
+}
+
 void CleanUpRasters()
 {
    register int i;
 
-   for (i = 0; i < MAXCHOICES; i++) XFreePixmap(mainDisplay, choicePixmap[i]);
+   CleanUpPatFill();
+   XFreePixmap(mainDisplay, rotateModePixmap[0]);
+   XFreePixmap(mainDisplay, rotateModePixmap[1]);
+   for (i = 0; i < MAXCHOICES-1; i++) XFreePixmap(mainDisplay, choicePixmap[i]);
    for (i = 0; i < MAXPATTERNS+1; i++) XFreePixmap(mainDisplay, patPixmap[i]);
    for (i = 0; i < MAXLINEWIDTHS; i++) {
       XFreePixmap(mainDisplay, defaultLineWidthPixmap[i]);
@@ -1216,3 +1319,129 @@ int UpgradePenFill(pen_or_fill)
       return pen_or_fill;
    }
 }
+
+unsigned char ByteReverse(byte)
+   unsigned char byte;
+{
+   int b=0;
+   unsigned char flip=0, mask=1;
+
+   for (b=0; b < 8; b++) {
+      flip <<= 1;
+      flip += byte&mask;
+      byte >>= 1;
+   }
+   return flip;
+}
+
+void GetPatternCellsize(fill_or_pen, pn_cell_w, pn_cell_h)
+   int fill_or_pen, *pn_cell_w, *pn_cell_h;
+{
+   int w=pat_w[fill_or_pen];
+   int bits_in_last_byte=(w & 0x07);
+   int bytes_per_row_out=0;
+
+   switch (bits_in_last_byte) {
+   case 1: bytes_per_row_out = (w+1); break;
+   case 2: bytes_per_row_out = ((w>>3)<<2)+1; break;
+   case 3: bytes_per_row_out = (w+3); break;
+   case 4: bytes_per_row_out = ((w>>3)<<1)+1; break;
+   case 5: bytes_per_row_out = (w+5); break;
+   case 6: bytes_per_row_out = ((w>>3)<<2)+3; break;
+   case 7: bytes_per_row_out = (w+7); break;
+   default: bytes_per_row_out = (w>>3); break;
+   }
+   if (pn_cell_w != NULL) *pn_cell_w = (bytes_per_row_out<<3);
+   if (pn_cell_h != NULL) *pn_cell_h = pat_h[fill_or_pen];
+}
+
+static
+void SetBit(buf, bit_offset, ival)
+   unsigned char *buf;
+   int bit_offset, ival;
+{
+   int byte_index=(bit_offset>>3);
+   int bit_in_byte=(7-(bit_offset&0x7));
+   unsigned char *psz=(&buf[byte_index]);
+
+   if (ival) {
+      *psz |= (((unsigned char)0x1) << bit_in_byte);
+   } else {
+      *psz &= (~(((unsigned char)0x1) << bit_in_byte));
+   }
+}
+
+static
+unsigned char GetBit(buf, bit_offset)
+   unsigned char *buf;
+   int bit_offset;
+{
+   int byte_index=(bit_offset>>3);
+   int bit_in_byte=(7-(bit_offset&0x7));
+   unsigned char uch=buf[byte_index];
+
+   return ((uch >> bit_in_byte) & ((unsigned char)0x1));
+}
+
+unsigned char *GetPatternBytes(fill_or_pen, pn_num_bytes, pn_need_to_free)
+   int fill_or_pen, *pn_num_bytes, *pn_need_to_free;
+{
+   int w=pat_w[fill_or_pen];
+   int h=pat_h[fill_or_pen];
+   int bits_in_last_byte=(w & 0x07);
+   int multiple_of_8=(bits_in_last_byte == 0);
+   int bytes_per_row_in=0, bytes_per_row_out=0, total_bytes=0;
+
+   bytes_per_row_in = (multiple_of_8 ? (w>>3) : (w>>3)+1);
+
+   switch (bits_in_last_byte) {
+   case 1: bytes_per_row_out = (w+1); break;
+   case 2: bytes_per_row_out = ((w>>3)<<2)+1; break;
+   case 3: bytes_per_row_out = (w+3); break;
+   case 4: bytes_per_row_out = ((w>>3)<<1)+1; break;
+   case 5: bytes_per_row_out = (w+5); break;
+   case 6: bytes_per_row_out = ((w>>3)<<2)+3; break;
+   case 7: bytes_per_row_out = (w+7); break;
+   default: bytes_per_row_out = (w>>3); break;
+   }
+   total_bytes = h * bytes_per_row_out;
+
+   if (pn_num_bytes != NULL) *pn_num_bytes = total_bytes;
+   if (!multiple_of_8) {
+      unsigned char *buf=(unsigned char *)malloc(total_bytes);
+      unsigned char *buf_in=patData[fill_or_pen], *buf_out=NULL;
+      int r=0, i=0;
+
+      if (buf == NULL) FailAllocMessage();
+      /* first reverse all the bits in buf_in */
+      for (i=0; i < bytes_per_row_in*h; i++) {
+         buf_in[i] = ByteReverse(buf_in[i]);
+      }
+      buf_out = buf;
+      for (r=0; r < h; r++) {
+         int c=0;
+
+         for (c=0; c < (bytes_per_row_out<<3); c++) {
+            int bit_offset_in=(c % w);
+
+            SetBit(buf_out, c, GetBit(buf_in, bit_offset_in));
+         }
+         buf_in += bytes_per_row_in;
+         buf_out += bytes_per_row_out;
+      }
+      /* reverse all the bits in buf_in back to what they were */
+      buf_in = patData[fill_or_pen];
+      for (i=0; i < bytes_per_row_in*h; i++) {
+         buf_in[i] = ByteReverse(buf_in[i]);
+      }
+      for (i=0; i < total_bytes; i++) {
+         buf[i] = ByteReverse(buf[i]);
+      }
+      if (pn_need_to_free != NULL) *pn_need_to_free = TRUE;
+      return buf;
+   }
+   if (pn_need_to_free != NULL) *pn_need_to_free = FALSE;
+
+   return patData[fill_or_pen];
+}
+
