@@ -1,7 +1,7 @@
 /*
  * Author:      William Chia-Wei Cheng (bill.cheng@acm.org)
  *
- * Copyright (C) 2001-2003, William Chia-Wei Cheng.
+ * Copyright (C) 2001-2009, William Chia-Wei Cheng.
  *
  * This file may be distributed under the terms of the Q Public License
  * as defined by Trolltech AS of Norway and appearing in the file
@@ -15,7 +15,7 @@
  * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *
- * @(#)$Header: /mm/src/tgif/v4/RCS/shortcut.c,v 4.93 2003/08/16 19:38:53 william Exp $
+ * @(#)$Header: /mm2/home/cvs/bc-src/tgif/shortcut.c,v 1.33 2009/09/28 22:39:48 william Exp $
  */
 
 #define _INCLUDE_FROM_SHORTCUT_C_
@@ -23,11 +23,13 @@
 #include "tgifdefs.h"
 #include "cmdids.h"
 
+#include "choice.e"
 #include "msg.e"
 #include "setup.e"
 #include "shortcut.e"
 #include "strtbl.e"
 #include "tangram2.e"
+#include "util.e"
 
 /*
  * Note:  the num_args field in shortCutXlateTbl is used by the shortcut
@@ -111,7 +113,7 @@ static struct ShortCutRec shortCutXlateTbl[] = {
    { ';', META_ONLY,  "CutBit/Pixmap()", 0, CMDID_CUTBITPIXMAP },
    { '_', META_ONLY,  "AbutHorizontal()", 0, CMDID_ABUTHORIZONTAL },
    { '|', META_ONLY,  "AbutVertical()", 0, CMDID_ABUTVERTICAL },
-   { '#', META_ONLY,  "BreakUpText()", 0, CMDID_BREAKUPTEXT },
+   { '#', META_ONLY,  "BreakUpText()", 1, CMDID_BREAKUPTEXT },
    { '^', META_ONLY,  "ScrollToOrig()", 0, CMDID_SCROLLTOORIG },
    { '@', META_ONLY,  "ToggleMoveMode()", 0, CMDID_TOGGLEMOVEMODE },
    { '$', META_ONLY,  "SelectVertexMode()", 0, CMDID_SELECTVERTEXMODE },
@@ -140,7 +142,7 @@ static struct ShortCutRec shortCutXlateTbl[] = {
    { 'p', CNTRL_META, "BoldItalic()", 0, CMDID_BOLDITALIC },
    { 'q', CNTRL_META, "DrawPolygon()", 0, CMDID_DRAWPOLYGON },
    { 'r', CNTRL_META, "Right()", 0, CMDID_RIGHT },
-   { 's', CNTRL_META, "SaveNew()", 0, CMDID_SAVENEW },
+   { 's', CNTRL_META, "SaveNew()", 1, CMDID_SAVENEW },
    { 't', CNTRL_META, "Italic()", 0, CMDID_ITALIC },
    { 'u', CNTRL_META, "UpdateSymbols()", 0, CMDID_UPDATESYMBOLS },
    { 'v', CNTRL_META, "DrawOval()", 0, CMDID_DRAWOVAL },
@@ -422,6 +424,39 @@ static struct ShortCutRec shortCutXlateTbl[] = {
    { '\0', 0, "ResetInputMethod()", 0, CMDID_RESETINMETHOD },
    { '\0', 0, "ImportPBMFile()", 0, CMDID_IMPORTPBMFILE },
    { '\0', 0, "ImportPGMFile()", 0, CMDID_IMPORTPGMFILE },
+   { '\0', 0, "EditIndexedAttrGroupInEditor()", 0,
+               CMDID_EDITATTRGROUPINEDITOR },
+   { '\0', 0,  "LinkExtJPEGFile()", 0, CMDID_LINKEXTJPEGFILE },
+   { '\0', 0, "NextSlide()", 0, CMDID_NEXTSLIDE },
+   { '\0', 0, "PrevSlide()", 0, CMDID_PREVSLIDE },
+   { '\0', 0, "SetObjectShadowColor()", 0, CMDID_SETOBJSHADOWCOLOR },
+   { '\0', 0, "SetObjectShadowOffsets()", 0, CMDID_SETOBJSHADOWOFFSETS },
+   { '\0', 0, "AddObjectShadow()", 0, CMDID_ADDOBJSHADOW },
+   { '\0', 0, "RemoveObjectShadow()", 0, CMDID_REMOVEOBJSHADOW },
+   { '\0', 0, "CopyDoubleByteString()", 0, CMDID_COPYUTF8 },
+   { '\0', 0, "PasteDoubleByteString()", 0, CMDID_PASTEUTF8 },
+   { '\0', 0, "ReduceToMobileWebSafeColors()", 0,
+               CMDID_REDUCETOMOBILEWEBCOLORS },
+   { '\0', 0, "CreatePixmapFromSelected()", 0, CMDID_CREATEPIXMAPFROMSEL },
+   { '\0', 0, "ToggleAutoRotatePivot()", 0, CMDID_TOGGLEROTATEPIVOT },
+   { '\0', 0, "SpecifyRotatePivot()", 0, CMDID_SPECIFYROTATEPIVOT },
+   { '\0', 0, "ResetRotatePivot()", 0, CMDID_RESETROTATEPIVOT },
+   { '\0', 0, "NextPolyRotationPivot()", 0, CMDID_NEXTPOLYROTATEPIVOT },
+   { '\0', 0, "MoveRotationPivotToArcCenter()", 0,
+               CMDID_MOVEROTATEPIVOTARCCNTR },
+   { '\0', 0, "MoveRotatePivotCenter()", 0, CMDID_MOVEROTATEPIVOTCENTER },
+   { '\0', 0, "MoveRotatePivotLeftTop()", 0, CMDID_MOVEROTATEPIVOTLT },
+   { '\0', 0, "MoveRotatePivotRightTop()", 0, CMDID_MOVEROTATEPIVOTRT },
+   { '\0', 0, "MoveRotatePivotLeftBottom()", 0, CMDID_MOVEROTATEPIVOTLB },
+   { '\0', 0, "MoveRotatePivotRightBottom()", 0, CMDID_MOVEROTATEPIVOTRB },
+   { '\0', 0, "MoveRotatePivotLeft()", 0, CMDID_MOVEROTATEPIVOTLEFT },
+   { '\0', 0, "MoveRotatePivotRight()", 0, CMDID_MOVEROTATEPIVOTRIGHT },
+   { '\0', 0, "MoveRotatePivotTop()", 0, CMDID_MOVEROTATEPIVOTTOP },
+   { '\0', 0, "MoveRotatePivotBottom()", 0, CMDID_MOVEROTATEPIVOTBOTTOM },
+   { '\0', 0, "ChooseRotatePivot()", 0, CMDID_CHOOSEROTATEPIVOT },
+   { '\0', 0, "ChooseColor()", 0, CMDID_COLORWHEEL },
+   { '\0', 0, "ExtendSegment()", 0, CMDID_EXTENDSEGMENT },
+   { '\0', 0, "ToggleTighterStructuredSplines()", 0, CMDID_TOGGLETIGHTERSPLINES },
    { '\0', 0, "", 0, 0 }
 };
 
@@ -429,22 +464,56 @@ static int maxShortCuts=0;
 static int shortCutIndex[256];
 static char *shortCutArgs[256];
 
+typedef struct tagFuncKeyShortCutInfo {
+   char code;
+   unsigned int state;
+   char *name;
+   char *arg;
+} FuncKeyShortCutInfo;
+
+static FuncKeyShortCutInfo funcKeyInfo[13];
+
 static int shortCutCmdIdIndex[MAXCMDIDS-CMDID_BASE];
 
+static
+int FunctionKeySpec(buf, pn_code, pn_index_return)
+   char *buf;
+   int *pn_code, *pn_index_return;
+{
+   int value=0;
+   char *colon_ptr=strchr(buf, ':');
+
+   if (colon_ptr == NULL) return FALSE;
+   *colon_ptr = '\0';
+   if (buf[0] != 'F' || sscanf(&buf[1], "%d", &value) != 1) {
+      *colon_ptr = ':';
+      return FALSE;
+   }
+   *colon_ptr++ = ':';
+   if (value < 1 || value > 12) return FALSE;
+
+   *pn_code = value;
+   *pn_index_return = (int)(colon_ptr-buf);
+
+   return TRUE;
+}
+
 void InitShortCut()
 {
-   int i, num_cmdids=MAXCMDIDS-CMDID_BASE;
-   char *c_ptr;
+   int i=0, num_cmdids=MAXCMDIDS-CMDID_BASE;
+   char *c_ptr=NULL;
 
    for (i=0; i < 256; i++) {
       shortCutIndex[i] = INVALID;
       shortCutArgs[i] = NULL;
    }
+   memset(funcKeyInfo, 0, 13*sizeof(FuncKeyShortCutInfo));
+
    maxShortCuts = 0;
    if ((c_ptr=XGetDefault(mainDisplay, TOOL_NAME, "ShortCuts")) != NULL) {
-      int j, precise_match, ok, len;
-      unsigned int code;
-      char *saved_c_ptr, *tmp_c_ptr, *buf;
+      int j=0, precise_match=0, ok=0, len=0;
+      unsigned int code=0;
+      char *saved_c_ptr=NULL, *tmp_c_ptr=NULL, *buf=NULL;
 
       len = strlen(c_ptr);
       if ((buf=(char *)malloc((len+1)*sizeof(char))) == NULL) {
@@ -473,6 +542,8 @@ void InitShortCut()
             if (*tmp_c_ptr == '\0') {
                ok = FALSE;
             } else {
+               int func_index=0;
+
                if (*tmp_c_ptr == '\n') {
                   *tmp_c_ptr = '\0';
                } else {
@@ -482,7 +553,7 @@ void InitShortCut()
                if (buf[0] != '\0' && buf[1] == ':' &&
                      ((code>0x20 && code<=0x7f) || (code>0xa0 && code<=0xff))) {
                   int found=TRUE;
-                  char *paren_ptr;
+                  char *paren_ptr=NULL;
    
                   if ((paren_ptr=strchr(&buf[2], '(')) == NULL) {
                      fprintf(stderr,
@@ -492,7 +563,8 @@ void InitShortCut()
                   }
                   len = paren_ptr-(&buf[2]);
                   for (j=0; *(shortCutXlateTbl[j].name) != '\0'; j++) {
-                     if (strncmp(shortCutXlateTbl[j].name, &buf[2], len) == 0) {
+                     if (strncmp(shortCutXlateTbl[j].name, &buf[2], len) == 0 &&
+                           shortCutXlateTbl[j].name[len] == '(') {
                         if (shortCutXlateTbl[j].num_args != 0) {
                            shortCutArgs[code] = (char*)malloc(
                                  (strlen(&buf[len+2])+1)*sizeof(char));
@@ -510,18 +582,16 @@ void InitShortCut()
                            if (*buf>='a' && *buf<='z') {
                               code = *buf-'a'+'A';
                               if (shortCutIndex[code] != INVALID) {
-                                 fprintf(stderr,
-                                       TgLoadString(STID_DUP_SHORTCUT_ENTRY_FOR_CHAR),
-                                       code);
+                                 fprintf(stderr, TgLoadString(
+                                       STID_DUP_SHORTCUT_ENTRY_FOR_CHAR), code);
                                  fprintf(stderr, "\n");
                               }
                               shortCutIndex[code] = j;
                            } else if (*buf>='A' && *buf<='Z') {
                               code = *buf-'A'+'a';
                               if (shortCutIndex[code] != INVALID) {
-                                 fprintf(stderr,
-                                       TgLoadString(STID_DUP_SHORTCUT_ENTRY_FOR_CHAR),
-                                       code);
+                                 fprintf(stderr, TgLoadString(
+                                       STID_DUP_SHORTCUT_ENTRY_FOR_CHAR), code);
                                  fprintf(stderr, "\n");
                               }
                               shortCutIndex[code] = j;
@@ -532,52 +602,97 @@ void InitShortCut()
                      }
                   }
                   if (*(shortCutXlateTbl[j].name) == '\0') {
-#ifdef _TANGRAM_II
-                     int num_args=0;
-                     int index=ValidTangram2CmdName(&buf[2], len, &num_args);
+                     if (cmdLineTgrm2) {
+                        int num_args=0;
+                        int index=ValidTangram2CmdName(&buf[2], len, &num_args);
 
-                     if (index != 0) {
-                        if (num_args != 0) {
-                           shortCutArgs[code] = (char*)malloc(
-                                 (strlen(&buf[len+2])+1)*sizeof(char));
-                           if (shortCutArgs[code] == NULL) FailAllocMessage();
-                           strcpy(shortCutArgs[code], ++paren_ptr);
-                        }
-                        if (shortCutIndex[code] != INVALID) {
-                           fprintf(stderr,
-                                 TgLoadString(STID_DUP_SHORTCUT_ENTRY_FOR_CHAR),
-                                 code);
-                           fprintf(stderr, "\n");
-                        }
-                        shortCutIndex[code] = index;
-                        if (!precise_match) {
-                           if (*buf>='a' && *buf<='z') {
-                              code = *buf-'a'+'A';
-                              if (shortCutIndex[code] != INVALID) {
-                                 fprintf(stderr,
-                                       TgLoadString(STID_DUP_SHORTCUT_ENTRY_FOR_CHAR),
-                                       code);
-                                 fprintf(stderr, "\n");
+                        if (index != 0) {
+                           if (num_args != 0) {
+                              shortCutArgs[code] = (char*)malloc(
+                                    (strlen(&buf[len+2])+1)*sizeof(char));
+                              if (shortCutArgs[code] == NULL) {
+                                 FailAllocMessage();
                               }
-                              shortCutIndex[code] = index;
-                           } else if (*buf>='A' && *buf<='Z') {
-                              code = *buf-'A'+'a';
-                              if (shortCutIndex[code] != INVALID) {
-                                 fprintf(stderr,
-                                       TgLoadString(STID_DUP_SHORTCUT_ENTRY_FOR_CHAR),
-                                       code);
-                                 fprintf(stderr, "\n");
-                              }
-                              shortCutIndex[code] = index;
+                              strcpy(shortCutArgs[code], ++paren_ptr);
                            }
+                           if (shortCutIndex[code] != INVALID) {
+                              fprintf(stderr, TgLoadString(
+                                    STID_DUP_SHORTCUT_ENTRY_FOR_CHAR), code);
+                              fprintf(stderr, "\n");
+                           }
+                           shortCutIndex[code] = index;
+                           if (!precise_match) {
+                              if (*buf>='a' && *buf<='z') {
+                                 code = *buf-'a'+'A';
+                                 if (shortCutIndex[code] != INVALID) {
+                                    fprintf(stderr, TgLoadString(
+                                          STID_DUP_SHORTCUT_ENTRY_FOR_CHAR),
+                                          code);
+                                    fprintf(stderr, "\n");
+                                 }
+                                 shortCutIndex[code] = index;
+                              } else if (*buf>='A' && *buf<='Z') {
+                                 code = *buf-'A'+'a';
+                                 if (shortCutIndex[code] != INVALID) {
+                                    fprintf(stderr, TgLoadString(
+                                          STID_DUP_SHORTCUT_ENTRY_FOR_CHAR),
+                                          code);
+                                    fprintf(stderr, "\n");
+                                 }
+                                 shortCutIndex[code] = index;
+                              }
+                           }
+                           maxShortCuts++;
+                        } else {
+                           found = FALSE;
                         }
-                        maxShortCuts++;
                      } else {
                         found = FALSE;
                      }
-#else /* ~_TANGRAM_II */
-                     found = FALSE;
-#endif /* _TANGRAM_II */
+                  }
+                  if (!found) {
+                     fprintf(stderr,
+                           TgLoadString(STID_INVALID_SHORTCUT_ENTRY_SKIP), buf);
+                     fprintf(stderr, "\n");
+                  }
+               } else if (FunctionKeySpec(buf, (int*)&code, &func_index)) {
+                  int found=FALSE;
+                  char *paren_ptr=NULL;
+   
+                  if ((paren_ptr=strchr(&buf[func_index], '(')) == NULL) {
+                     fprintf(stderr,
+                           TgLoadString(STID_INVALID_SHORTCUT_ENTRY_SKIP), buf);
+                     fprintf(stderr, "\n");
+                     continue;
+                  }
+                  len = paren_ptr-(&buf[func_index]);
+                  for (j=0; *(shortCutXlateTbl[j].name) != '\0'; j++) {
+                     if (strncmp(shortCutXlateTbl[j].name, &buf[func_index],
+                           len) == 0 && shortCutXlateTbl[j].name[len] == '(') {
+                        if (funcKeyInfo[code].name != NULL) {
+                           fprintf(stderr,
+                                 TgLoadString(STID_DUP_SHORTCUT_FUNCTION_KEY),
+                                 code);
+                           fprintf(stderr, "\n");
+                        }
+                        funcKeyInfo[code].code = shortCutXlateTbl[j].code;
+                        funcKeyInfo[code].state = shortCutXlateTbl[j].state;
+                        funcKeyInfo[code].name =
+                              UtilStrDup(shortCutXlateTbl[j].name);
+                        if (funcKeyInfo[code].name == NULL) {
+                           FailAllocMessage();
+                        }
+                        if (shortCutXlateTbl[j].num_args != 0) {
+                           funcKeyInfo[code].arg = (char*)malloc(
+                                 (strlen(&buf[len+func_index])+1)*sizeof(char));
+                           if (funcKeyInfo[code].arg == NULL) {
+                              FailAllocMessage();
+                           }
+                           strcpy(funcKeyInfo[code].arg, ++paren_ptr);
+                        }
+                        found = TRUE;
+                        break;
+                     }
                   }
                   if (!found) {
                      fprintf(stderr,
@@ -633,17 +748,20 @@ void CleanUpShortCut()
          free(shortCutArgs[i]);
       }
    }
+   for (i=0; i < 13; i++) {
+      UtilFree(funcKeyInfo[i].name);
+      UtilFree(funcKeyInfo[i].arg);
+   }
 }
 
-int FetchShortCut(c, code, state, name, args)
-   int c;
+int FetchShortCut(c, code, state, name, args, args_sz)
+   int c, args_sz;
    char *code, **name, *args;
    unsigned int *state;
 {
    if (shortCutIndex[c] == INVALID) return FALSE;
 
-#ifdef _TANGRAM_II
-   if ((shortCutIndex[c] & CMDID_TANGRAM2_BASE) != 0) {
+   if (cmdLineTgrm2 && (shortCutIndex[c] & CMDID_TANGRAM2_BASE) != 0) {
       /* for tangram-II, just do it */
       int num_args=0;
 
@@ -657,7 +775,6 @@ int FetchShortCut(c, code, state, name, args)
 
       return FALSE;
    }
-#endif /* _TANGRAM_II */
    *code = shortCutXlateTbl[shortCutIndex[c]].code;
    *state = shortCutXlateTbl[shortCutIndex[c]].state;
    *name = shortCutXlateTbl[shortCutIndex[c]].name;
@@ -669,6 +786,49 @@ int FetchShortCut(c, code, state, name, args)
    return TRUE;
 }
 
+int FetchFuncKeyShortCut(c, code, state, name, args, args_sz)
+   int c, args_sz;
+   char *code, **name, *args;
+   unsigned int *state;
+{
+   int index=0;
+   char buf[MAXSTRING];
+
+   switch (c) {
+   case XK_F1: index=1; break;
+   case XK_F2: index=2; break;
+   case XK_F3: index=3; break;
+   case XK_F4: index=4; break;
+   case XK_F5: index=5; break;
+   case XK_F6: index=6; break;
+   case XK_F7: index=7; break;
+   case XK_F8: index=8; break;
+   case XK_F9: index=9; break;
+   case XK_F10: index=10; break;
+   case XK_F11: index=11; break;
+   case XK_F12: index=12; break;
+   default: return FALSE;
+   }
+   if (curChoice == DRAWTEXT) {
+      sprintf(buf, "DrawTextFuncKey_F%1d", index);
+      if (XGetDefault(mainDisplay, TOOL_NAME, buf) != NULL) {
+         return FALSE;
+      }
+   }
+   if (funcKeyInfo[index].name == NULL) return FALSE;
+
+   *code = funcKeyInfo[index].code;
+   *state = funcKeyInfo[index].state;
+   *name = funcKeyInfo[index].name;
+
+   if (funcKeyInfo[index].arg == NULL) {
+      *args = '\0';
+   } else {
+      strcpy(args, funcKeyInfo[index].arg);
+   }
+   return TRUE;
+}
+
 int FetchCmdById(nCmdId, code, state, name, args)
    int nCmdId;
    char *code, **name, *args;
